/* Written by Dmitrij Schlesinger (2012),
 * http://wwwpub.zih.tu-dresden.de/~ds24/rtsegm/rtsegm.html
*/

#include <iostream>
using namespace std;
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "ed_rand.h"
#include "rtsegm_gauss.h"

// constructor -- remember parameters, allocate arrays, initialize
Rtsegm_gauss::Rtsegm_gauss(int height, int width, int ng, double step)
    : ed_thread() {
    this->height=height;
    this->width=width;
    this->ng=ng;
    this->step=step;
    gnoise=1.e-10;
    img=0;

    m_r=new double[ng];
    m_g=new double[ng];
    m_b=new double[ng];

    for( int ig=0; ig<ng; ig++ )
        m_r[ig]=m_g[ig]=m_b[ig]=128.;
    sigma=10000.;

    weights=new double[ng*2];
    for( int i=0; i<ng*2; i++ ) weights[i]=log(1./ng);

    glab=new int[height*width];
    ed_rand *generator=new ed_rand((unsigned int)time(0));
    for( int i=0; i<height*width; i++ ) glab[i]=generator->rand()%ng;
    delete generator;

    stop_me=false;
    it_count=0;

    kr=new double[ng*ng];
    kg=new double[ng*ng];
    kb=new double[ng*ng];
    t=new double[ng*ng];
    do_coeff();

    acc_r=new double[ng];
    acc_g=new double[ng];
    acc_b=new double[ng];
    acc_weights=new double[ng*2];

    // lookup table
    table=new unsigned short[0x10000]; // 128k
    for( int i=0; i<0x10000; i++ ) table[i]=(unsigned short)(exp(-i/1000.)*0xFFFF+0.5);
}

// destructor -- clean all
Rtsegm_gauss::~Rtsegm_gauss() {
    delete m_r;
    delete m_g;
    delete m_b;
    delete weights;
    delete glab;
    delete kr;
    delete kg;
    delete kb;
    delete t;
    delete acc_r;
    delete acc_g;
    delete acc_b;
    delete acc_weights;
    delete table;
}

// the entry point for separate thread -- Metropolis Sampling for Gaussians
void Rtsegm_gauss::run() {
    // construct own random generator to be thread-safe
    ed_rand *generator=new ed_rand((unsigned int)time(0)+54321);
    stop_me=false;

    // the main loop
    while( !stop_me ) {
        // initialize statistics with a small portion of "noise"
        // imagine, we have an (image,segmentation)-pair, with random segmentation,
        // uniform weights and current Gaussian parameters (means and dispersion)
        for( int ig=0; ig<ng; ig++ ) {
            acc_r[ig]=m_r[ig]*width*height*gnoise/ng;
            acc_g[ig]=m_g[ig]*width*height*gnoise/ng;
            acc_b[ig]=m_b[ig]*width*height*gnoise/ng;
            acc_weights[ig*2]=acc_weights[ig*2+1]=width*height*gnoise/ng/2.;
        }
        acc_sigma=sigma*width*height*gnoise*3.;

        // scan over the image
        int hw=height*width;
        for( int ii=0, i=0; i<hw; i++ ) {
            int red=img[ii++];
            int green=img[ii++];
            int blue=img[ii++];
            int oldg=glab[i];
            int newg=generator->rand()%ng;
            // if the new number is the same as the current one, skip the sampling
            if( newg!=oldg ) {
                int gg=newg*ng+oldg;
                // the necessary energy difference is a scalar product with precomputed coefficients
                double diff=red*kr[gg]+green*kg[gg]+blue*kb[gg]+t[gg]+weights[newg*2+lab[i]]-weights[oldg*2+lab[i]];
                if( diff>0. ) glab[i]=newg; // the new energy is better -- accept the new Gaussian
                else { // otherwise sample
                    diff=-diff*1000.+0.5;
                    unsigned int index;
                    if( diff>=(double)0x10000 ) index=0xFFFF;
                    else index=(int)diff;
                    if( generator->rand() < table[index] ) glab[i]=newg;
                }
            }
            int ig=glab[i];

            // accumulate statistics
            acc_r[ig]+=red;
            acc_g[ig]+=green;
            acc_b[ig]+=blue;
            acc_weights[ig*2+lab[i]]+=1.;

            double diff=red-m_r[ig];
            acc_sigma+=diff*diff;
            diff=green-m_g[ig];
            acc_sigma+=diff*diff;
            diff=blue-m_b[ig];
            acc_sigma+=diff*diff;
        } // end of the scan over the image

        // update parameters
        double sum0=0.;
        double sum1=0.;
        for( int ig=0; ig<ng; ig++ ) {
            m_r[ig]=m_r[ig]*(1.-step)+acc_r[ig]/(acc_weights[ig*2]+acc_weights[ig*2+1])*step;
            m_g[ig]=m_g[ig]*(1.-step)+acc_g[ig]/(acc_weights[ig*2]+acc_weights[ig*2+1])*step;
            m_b[ig]=m_b[ig]*(1.-step)+acc_b[ig]/(acc_weights[ig*2]+acc_weights[ig*2+1])*step;
            sum0+=acc_weights[ig*2];
            sum1+=acc_weights[ig*2+1];
        }

        for( int ig=0; ig<ng; ig++ ) {
            weights[ig*2]=log(exp(weights[ig*2])*(1.-step)+acc_weights[ig*2]/sum0*step);
            weights[ig*2+1]=log(exp(weights[ig*2+1])*(1.-step)+acc_weights[ig*2+1]/sum1*step);
        }

        sigma=sigma*(1.-step)+2.*acc_sigma/3./(sum0+sum1)*step;

        // recalculate coefficients
        do_coeff();

        it_count++;
    }

    stop_me=false;
    delete generator;
}

// recalculate coefficients for the scalar products in "::run()"
void Rtsegm_gauss::do_coeff() {
    for( int g1=0; g1<ng; g1++ ) {
        for( int g2=0; g2<ng; g2++ ) {
            kr[g1*ng+g2]=(2.*m_r[g1]-2.*m_r[g2])/sigma;
            kg[g1*ng+g2]=(2.*m_g[g1]-2.*m_g[g2])/sigma;
            kb[g1*ng+g2]=(2.*m_b[g1]-2.*m_b[g2])/sigma;
            t[g1*ng+g2]=-(m_r[g1]*m_r[g1]-m_r[g2]*m_r[g2]+m_g[g1]*m_g[g1]-m_g[g2]*m_g[g2]+m_b[g1]*m_b[g1]-m_b[g2]*m_b[g2])/sigma;
        }
    }
}

// reset the parameters, re-initialise "glab" randomly
void Rtsegm_gauss::reset() {
    for( int ig=0; ig<ng; ig++ )
        m_r[ig]=m_g[ig]=m_b[ig]=128.;
    sigma=10000.;
    for( int i=0; i<ng*2; i++ ) weights[i]=log(1./ng);
    do_coeff();

    ed_rand *generator=new ed_rand((unsigned int)time(0));
    for( int i=0; i<height*width; i++ ) glab[i]=generator->rand()%ng;
    delete generator;

    stop_me=false;
}
