/* Written by Dmitrij Schlesinger (2012-2013),
 * http://wwwpub.zih.tu-dresden.de/~ds24/rtsegm/rtsegm.html
 *
 * See paper "A Continuous Shape Prior for MRF-based Segmentation"
 * at EMMCVPR 2013 Conference for details.
 * 
 * Note: this is not the program, with which the results in the
 * paper were obtained. This is a real-time demostration only
 * (it was presented during the tals).
 * 
 * The code is heavily based (but not compatible!!!) on the previous
 * version -- see "A Real-time MRF Based Approach for Binary Segmentation"
 * at DAGM-OAGM 2012 Conference for details.
*/

// standard stuff
#include <stdio.h>
#include <iostream>
using namespace std;
#include <stdlib.h>
#include <time.h>

// input/output
#include "d_inout.h"

// rtsegm stuff
#include "rtsegm_gauss.h"
#include "rtsegm_segm.h"
#include "rtsegm_shape.h"

// visualize marginals, comment it to visualize the generated labellings
// #define VIS_MARG

int main() {
    // Parameters:

    // Input -- camera, sizes etc.
    int mode=0;                 // input mode: 0 -- webcam, 1 -- video, 2 -- still image
    int camera=0;               // for mode=0
    const char *filename="..."; // for mode=1,2
    int height=120;	            // 0 means "use native (do not change)"
    int width=160;              // 0 means "use native (do not change)"
    bool flip=true;             // 0 -- do not flip, 1 -- mirror (for laptop webcams :-))
    int fps=0;                  // 0 means "try maximal" (helpful for webcams)
    int delay;		            // waiting time -- "almost" 1000/fps

    // modell (fixed parameter)
    // defaults are tuned for my working place:
    // quite light background, laptop webcam (30 fps), my face as the foreground,
    // quite fast computer (i7, quad-core, 2.8 GHz)
    
    int ngauss=16;         // number of Gaussians
    double potts=1.3;	   // Potts parameter (>0, additive)

    // learning speeds
    double stepgauss=0.1; // learning speed for Gaussians (0...1), the more the faster
    double stepmarg=0.1;  // speed for marginals accumulation, 0.01 -- slow, 0.1 -- fast
    double step_m=0.1;    // shape related gradient steps (see code in rtsegm_shape.cpp)
    double step_g=0.5;    //
    double step_d=0.1;    //
	double beta=1.2;      // the constant for the approximation of the prior marginals (Eq.7 in the EMMCVPR paper)

    // Input -- Webcam/Video/Image
    d_in *din=new d_in(mode,camera,(char *)filename,height,width,flip);
    height=din->height;
    width=din->width;

    if( fps ) delay=(int)(1000./fps+0.5);
    else delay=1;

    // array for the output image
    unsigned char *img_out=new unsigned char[width*height*3*4];

    // Output -- window
    d_out *dout=new d_out(0,(char *)"rtsegm",height*2,width*2,img_out);
    // Output -- video (no recording at the beginning)
    d_out *record=0;
    // Output -- screenshot
    d_out *screenshot;

    // Metropolis Sampler for Gaussians
    Rtsegm_gauss *gauss=new Rtsegm_gauss(height,width,ngauss,stepgauss);

    // Gibbs Sampler for Segmentations
    Rtsegm_segm *segm=new Rtsegm_segm(height,width,potts,stepmarg);

    // Shape optimizer
    Rtsegm_shape *shape=new Rtsegm_shape(height,width,step_m,step_g,step_d,beta);

    // give access to the fields (see Fig.2 in the EMMCVPR-paper)
    // Rtsegm_gauss should observe the image
    gauss->img=din->img;
    // Rtsegm_gauss should observe the segmentation
    gauss->lab=segm->lab;
    // Rtsegm_segm should observe the field of Gaussian numbers
    segm->glab=gauss->glab;
    // Rtsegm_segm should observe the Gaussian weights
    segm->weights=gauss->weights;
    // Rtsegm_segm should observe the shape unary potentials
    segm->qsh=shape->qsh;
    // Rtsegm_shape should observe the segmentation
    shape->lab=segm->lab;

    // start
    gauss->start();
    segm->start();
    shape->start();

    // variables to watch real camera speed (frames per seconds)
    // and samplers/optimizer speeds (iterations per frame)
    int time0, frame_count;
    int countgprev, countgnext, countsprev, countsnext, countshprev, countshnext;
    double real_fps, gipf, sipf, shipf;

    time0=(int)time(0);
    frame_count=countgprev=countsprev=countshprev=0;
    if( fps ) real_fps=fps;
    else real_fps=10.;
    gipf=sipf=shipf=0.;

    // show some help
    cerr << endl << "     h -- help";
    cerr << endl << "     i -- info";
    cerr << endl << "     c -- start/stop capture";
    cerr << endl << "     s -- screenshot";
    cerr << endl << "     r -- reset";
    cerr << endl << "     p -- pause";
    cerr << endl << "   a/z -- potts +/-";
    cerr << endl << "     q -- shape on/off";
    cerr << endl << "   esc -- quit";
    cerr << endl << endl;

    // main loop
    bool flag_exit=false;
    while( !flag_exit ) {
        // wait a bit
        int key = din->key(delay);

        // watch real camera speed
        int time1=(int)time(0);
        if( time1>time0 ) {
            real_fps=real_fps*0.1+0.9*frame_count/(time1-time0);
            time0=time1;
            frame_count=0;
        }

        // watch sampler speed
        countgnext=gauss->it_count;
        countsnext=segm->it_count;
        countshnext=shape->it_count;
        gipf=gipf*0.99+0.01*(countgnext-countgprev);
        sipf=sipf*0.99+0.01*(countsnext-countsprev);
        shipf=shipf*0.99+0.01*(countshnext-countshprev);
        countgprev=countgnext;
        countsprev=countsnext;
        countshprev=countshnext;

        // what for a key
        if( key>0 ) {
            switch (key) {
            case 27: // escape -- exit
                flag_exit = true;
                break;
            case 'h': // show help
                cerr << endl << "     h -- help";
                cerr << endl << "     i -- info";
                cerr << endl << "     c -- start/stop capture";
                cerr << endl << "     s -- screenshot";
                cerr << endl << "     r -- reset";
                cerr << endl << "     p -- pause";
                cerr << endl << "   a/z -- potts +/-";
				cerr << endl << "     q -- shape on/off";
				cerr << endl << "   esc -- quit";
                cerr << endl << endl;
                break;
            case 'i': // show some info
                cerr << endl << height << "x" << width << ", " << real_fps << "fps, g:"
                     << gipf << "/" << gauss->step << ", s:"
                     << sipf << "/" << segm->step << " sh:"
                     << shipf << ", speed index: "
                     << real_fps*(gipf+sipf+shipf);

                cerr << endl << "Potts: " << potts;

                cerr << endl << "Sigma: " << gauss->sigma;
                
                cerr << endl << "Shape: [" << shape->min_s << "," << shape->max_s << "]";

                cerr << endl << "Weights:" << endl;
                for( int g=0; g<ngauss; g++ )
                    cerr << " " << gauss->weights[g*2];
                cerr << endl;
                for( int g=0; g<ngauss; g++ )
                    cerr << " " << gauss->weights[g*2+1];
                cerr << endl << endl;
                break;
            case 'c': // start/stop recording
                if( record ) {
                    cerr << endl << "record finished" << endl;
                    delete record;
                    record=0;
                } else {
                    // Attention!!! The video is created with 30 fps -- the real speed is ignored
                    record=new d_out(1,(char *)"rtsegm.avi",height*2,width*2,img_out);
                    cerr << endl << "record started -- \"rtsegm.avi\"" << endl;
                }
                break;
            case 's': // take screenshot
                screenshot=new d_out(2,(char *)"rtsegm.png",height*2,width*2,img_out);
                screenshot->out();
                delete screenshot;
                cerr << endl << "screenshot done -- \"rtsegm.png\"" << endl;
                break;
            case 'r': // reset
                segm->stop_me=true;
                gauss->stop_me=true;
                shape->stop_me=true;
                segm->wait();
                gauss->wait();
                shape->wait();
                segm->reset();
                gauss->reset();
                shape->reset();
                gauss->start();
                segm->start();
                shape->start();
                cerr << endl << "reset" << endl;
                break;
            case 'p': // pause
                segm->stop_me=true;
                gauss->stop_me=true;
                shape->stop_me=true;
                segm->wait();
                gauss->wait();
                shape->wait();
                cerr << endl << "pause, press any key to resume ... ";
                din->key(-1);
                gauss->start();
                segm->start();
                shape->start();
                cerr << "resumed" << endl;
                break;
            case 'q': // switch the shape on/off
                segm->stop_me=true;
                gauss->stop_me=true;
                shape->stop_me=true;
                segm->wait();
                gauss->wait();
                shape->wait();
                if( segm->qsh ) {
                    segm->qsh=0;
                    cerr << "shape is off" << endl;
                }
                else {
                    segm->qsh=shape->qsh;
                    cerr << "shape is on" << endl;
                }
                gauss->start();
                segm->start();
                shape->start();
                break;
            case 'a': // increase potts
                potts+=0.1;
                segm->potts=potts;
                cerr << endl << "potts=" << potts << endl;
                break;
            case 'z': // decrease potts
                potts-=0.1;
                if( potts<0.1 ) potts=0.1;
                segm->potts=potts;
                cerr << endl << "potts=" << potts << endl;
                break;
                break;
            default:
                break;
            }
        }

        // get the new frame
        if( !din->in() ) break; // 0 -- if there are no more frames in a video

        // prepare image for the visualization
        unsigned char *psrc=din->img;
        for( int ii=0, i=0; i<height; i++ ) {
            // input image -- left top
            unsigned char *pi=img_out+i*width*3*2;
            // Gaussians -- left bottom
            unsigned char *pg=pi+height*width*3*2;
            // shape -- right top
            unsigned char *pl=pi+width*3;
            // marginals -- right bottom
            unsigned char *pm=pl+height*width*3*2;
            for( int j=0; j<width; j++,ii++ ) {
                // image
                *(pi++)=*(psrc++);
                *(pi++)=*(psrc++);
                *(pi++)=*(psrc++);
                // gaussians
                int ig=gauss->glab[ii];
                *(pg++)=(unsigned char)(gauss->m_r[ig]+0.5);
                *(pg++)=(unsigned char)(gauss->m_g[ig]+0.5);
                *(pg++)=(unsigned char)(gauss->m_b[ig]+0.5);
                // shape
                unsigned char gray=(shape->qsh[ii]-shape->min_s)*255./(shape->max_s-shape->min_s);
                unsigned char r=gray, g=gray, b=gray;
                if( ii/width!=0 && ii/width!=height-1 && ii%width!=0 && ii%width!=width-1 ) {
                    // check for zero-level
                    int center=(shape->qsh[ii]>0)?1.:-1.;
                    int left=(shape->qsh[ii-1]>0)?1.:-1.;
                    int right=(shape->qsh[ii+1]>0)?1.:-1.;
                    int top=(shape->qsh[ii-width]>0)?1.:-1.;
                    int bottom=(shape->qsh[ii+width]>0)?1.:-1.;
                    // left
                    if( center*left<0 || center*right<0 || center*top<0 || center*bottom<0 ) {
                        r=0; g=0; b=255;
                    }
                }
                *(pl++)=r;
                *(pl++)=g;
                *(pl++)=b;
                // marginals/labelling
#ifdef VIS_MARG				
                gray=(unsigned char)(segm->marg[ii]*255.+0.5);
#else
                gray=(unsigned char)(segm->lab[ii]*255.+0.5);
#endif
                *(pm++)=gray;
                *(pm++)=gray;
                *(pm++)=gray;
            }
        }
        
        // show in window
        dout->out();

        // record the frame (if necessary)
        if( record ) record->out();

        frame_count++;
    } // end of the main loop

    // stop recording (if any)
    if( record ) delete record;

    // stop all processes
    segm->stop_me=true;
    gauss->stop_me=true;
    shape->stop_me=true;
    segm->wait();
    gauss->wait();
    shape->wait();

    // clean
    delete gauss;
    delete segm;
    delete shape;
    delete din;
    delete dout;
    delete img_out;
    exit(0);
}
