/* Written by Dmitrij Schlesinger (2012-2013),
 * http://wwwpub.zih.tu-dresden.de/~ds24/rtsegm/rtsegm.html
*/

#include <math.h>
#include <iostream>
using namespace std;
#include <stdlib.h>
#include <time.h>
#include <cstring>
#include "rtsegm_shape.h"

// activate if the shape should be initialized by by an allipse in the middle
#define INIT_EL

// constructor -- remember parameters, allocate arrays, initialize
Rtsegm_shape::Rtsegm_shape(int height, int width,
						   double step_m, double step_g, double step_d, double beta)
    : ed_thread() {
    this->height=height;
    this->width=width;
	this->step_m=step_m;
	this->step_g=step_g;
	this->step_d=step_d;
	this->beta=beta;

    stop_me=false;
    it_count=0;

    qsh=new double[height*width];

#ifdef INIT_EL    
    double x0=width/2.;
    double y0=height/2.;
    double srad=width*1./4.;
    if( height<width ) srad=height*1./4.;
    double level=0.1;
    min_s=100000000.;
    max_s=-100000000.;
    for( int i=0; i<width*height; i++ ) {
        double x=i%width;
        double y=i/width;
        double dist=(y-y0)*(y-y0)+(x-x0)*(x-x0);
        qsh[i]=level*(1.-dist/(srad*srad));
        if( qsh[i]<min_s ) min_s=qsh[i];
        if( qsh[i]>max_s ) max_s=qsh[i];
    }
#else
    for( int i=0; i<height*width; i++ ) qsh[i]=0.;
    min_s=-0.001;
    max_s=0.001;
#endif
    tmp_d=new double[height*width];
    tmp_g=new double[height*width];

#ifdef USE_LOOKUP	
	// lookup table
	lt=new double[1000];
	for( int i=0; i<1000; i++ ) lt[i]=exp(i/100.)/(1.+exp(i/100.));
#endif
}

// destructor -- clean
Rtsegm_shape::~Rtsegm_shape() {
    delete qsh;
    delete tmp_g;
    delete tmp_d;
#ifdef USE_LOOKUP	
	delete lt;
#endif
}

// the entry point for separate thread -- the shape optimizer
void Rtsegm_shape::run() {
    stop_me=false;

    // the main loop
    while( !stop_me ) {
        // do convolutions (compute the gradient parts)
        filter_g(qsh,tmp_g);
        filter_d(qsh,tmp_d);

        // gradient step
        double newmin=1000000000.;
        double newmax=-1000000000.;
        for( int i=0; i<height*width; i++ ) {
		  // the Likelihood part
#ifdef USE_LOOKUP		  
		  int index;
		  double margsh;
		  if( qsh[i]>0. ) {
			index=(int)(beta*qsh[i]*100.+0.5);
			if( index<1000 ) margsh=lt[index];
			else margsh=1.;
		  }
		  else {
			index=(int)(-beta*qsh[i]*100.+0.5);
			if( index<1000 ) margsh=1.-lt[index];
			else margsh=0.;
		  }
#else
          double margsh=exp(beta*qsh[i])/(1.+exp(beta*qsh[i]));
#endif
          qsh[i]+=step_m*(lab[i]-margsh)+step_g*(tmp_g[i]-qsh[i])-step_d*tmp_d[i];
          if( qsh[i]<newmin ) newmin=qsh[i];
          if( qsh[i]>newmax ) newmax=qsh[i];
        }
        min_s=newmin;
        max_s=newmax;

        it_count++;
    }
    stop_me=false;
}

// reset the shape
void Rtsegm_shape::reset() {
    stop_me=false;

#ifdef INIT_EL    
    double x0=width/2.;
    double y0=height/2.;
    double srad=width*1./4.;
    if( height<width ) srad=height*1./4.;
    double level=1.;
    min_s=100000000.;
    max_s=-100000000.;
    for( int i=0; i<width*height; i++ ) {
        double x=i%width;
        double y=i/width;
        double dist=(y-y0)*(y-y0)+(x-x0)*(x-x0);
        qsh[i]=level*(1.-dist/(srad*srad));
        if( qsh[i]<min_s ) min_s=qsh[i];
        if( qsh[i]>max_s ) max_s=qsh[i];
    }
#else
    for( int i=0; i<height*width; i++ ) qsh[i]=0.;
    min_s=-0.001;
    max_s=0.001;
#endif
}

void Rtsegm_shape::filter_g(double *src, double *dest) {
  double *tmp1=new double[width*height];
  
  // horizontal
  for( int i=0, ii=0; i<height; i++ ) {
	tmp1[ii]=(src[ii]*6.+src[ii+1]*4.+src[ii+2])/11.;
	ii++;
	tmp1[ii]=(src[ii-1]*4.+src[ii]*6.+src[ii+1]*4.+src[ii+2])/15.;
	ii++;
	for( int j=0; j<width-4; j++, ii++ )
	  tmp1[ii]=(src[ii-2]+src[ii-1]*4.+src[ii]*6.+src[ii+1]*4.+src[ii+2])/16.;
	tmp1[ii]=(src[ii-2]+src[ii-1]*4.+src[ii]*6.+src[ii+1]*4.)/15.;
	ii++;
	tmp1[ii]=(src[ii-2]+src[ii-1]*4.+src[ii]*6.)/11.;
	ii++;
  }
  
  // vertical
  int ii;
  for( ii=0; ii<width; ii++ )
	dest[ii]=(tmp1[ii]*6.+tmp1[ii+width]*4.+tmp1[ii+2*width])/11.;
  for( ; ii<2*width; ii++ )
	dest[ii]=(tmp1[ii-width]*4.+tmp1[ii]*6.+tmp1[ii+width]*4.+tmp1[ii+2*width])/15.;
  for( ; ii<(height-2)*width; ii++ )
	dest[ii]=(tmp1[ii-2*width]+tmp1[ii-width]*4.+tmp1[ii]*6.+tmp1[ii+width]*4.+tmp1[ii+2*width])/16.;
  for( ; ii<(height-1)*width; ii++ )
	dest[ii]=(tmp1[ii-2*width]+tmp1[ii-width]*4.+tmp1[ii]*6.+tmp1[ii+width]*4.)/15.;
  for( ; ii<height*width; ii++ )
	dest[ii]=(tmp1[ii-2*width]+tmp1[ii-width]*4.+tmp1[ii]*6.)/11.;
  
  delete tmp1;
}

void Rtsegm_shape::filter_d(double *src, double *dest) {
    double *tmp1=new double[width*height];

	// the first and the last rows
	for( int j=0, jj=(height-1)*width; j<width; j++, jj++ ) tmp1[j]=tmp1[jj]=0.;
	
	// vertical -1,2,-1 convolution
	for( int ii=width; ii<(height-1)*width; ii++ ) {
		tmp1[ii]=2.*src[ii]-src[ii-width]-src[ii+width];
	}
	
	// horizontal -1,2,-1 convolution (add)
	for( int i=0, ii=0; i<height; i++ ) {
	  ii++;
	  for( int j=1; j<width-1; j++, ii++ )
		tmp1[ii]=+2.*src[ii]-src[ii-1]-src[ii+1];
	  ii++;
	}

    for( int i=0, ii=0; i<height; i++ ) {
        for( int j=0; j<width; j++, ii++ ) {
            double xm1=tmp1[ii];
            if( j>0 ) xm1=tmp1[ii-1];
            double xp1=tmp1[ii];
            if( j<width-1 ) xp1=tmp1[ii+1];

            double ym1=tmp1[ii];
            if( i>0 ) ym1=tmp1[ii-width];
            double yp1=tmp1[ii];
            if( i<height-1 ) yp1=tmp1[ii+width];

            dest[ii]=-(4.*tmp1[ii]-xm1-xp1-ym1-yp1)/20.;
		}
    }

    delete tmp1;
}
